from __future__ import annotations  # for python 3.8


class Version:
    as_str: str = None
    as_ags4_str: str = None
    as_int: int = None
    family: str = None
    family_as_int: int = None
    major: str = None
    minor: str = None
    improv: str = None
    patch: str = None



from __future__ import annotations  # for python 3.8

from typing import cast

from PyQt6 import QtCore
from PyQt6.QtWidgets import QTreeWidget, QWidget, QAbstractScrollArea, QFrame, QTreeWidgetItem

from agstoolbox.at_tasks import do_update_tools_downloads, do_update_tools_unmanaged, \
    do_update_tools_managed
from agstoolbox.core.ags.ags_editor import LocalAgsEditor
from agstoolbox.core.ags.ags_local_run import ags_editor_load, ags_editor_build
from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.version.version import Version
from agstoolbox.wdgts.at_tree_item_tool import TreeItemTool_Header, ToolType, \
    TreeItemTool_Download, TreeItemTool_ExternallyInstalled, TreeItemTool_Managed, TreeItemTool


class ToolsTree(QTreeWidget):
    tool_update_downloads_task = None
    tool_update_unmanaged_task = None
    tool_update_managed_task = None
    header_managed = None
    header_download = None
    header_unmanaged = None
    managed_editors_list: list[LocalAgsEditor] = None
    unmanaged_editors_list: list[LocalAgsEditor] = None

    def __init__(self, parent: QWidget = None):
        QTreeWidget.__init__(self, parent)

        self.setHeaderHidden(True)
        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustIgnored)
        self.setObjectName("treeTools")
        self.setFrameStyle(QFrame.Shape.NoFrame)
        self.clear()
        self.header_managed = TreeItemTool_Header(
            "Managed", ToolType.MANAGED_TOOL_HEADER)
        self.header_unmanaged = TreeItemTool_Header(
            "Externally Installed", ToolType.EXTERNALLY_INSTALLED_TOOL_HEADER)
        self.header_download = TreeItemTool_Header(
            "Available for Download", ToolType.AVAILABLE_TO_DOWNLOAD_HEADER)
        self.addTopLevelItem(self.header_managed)
        self.addTopLevelItem(self.header_unmanaged)
        self.addTopLevelItem(self.header_download)
        self.setRootIsDecorated(False)
        self.setIndentation(0)

        # make headers expand or recover in a single click
        self.clicked.connect(self.toggleExpandSlot)
        self.itemExpanded.connect(self.itemIsExpanded)
        self.itemCollapsed.connect(self.itemIsCollapsed)

        self.header_managed.setExpanded(True)
        self.header_unmanaged.setExpanded(True)
        self.header_download.setExpanded(True)
        self.header_managed.setExpanded(False)
        self.header_unmanaged.setExpanded(False)

    def itemIsExpanded(self, itm: QTreeWidgetItem):
        itm.setText(0, "- " + itm.whatsThis(0))

    def itemIsCollapsed(self, itm: QTreeWidgetItem):
        itm.setText(0, "+ " + itm.whatsThis(0))

    def toggleExpandSlot(self, i):
        self.setExpanded(i, not self.isExpanded(i))

    ###############################################################################################
    # Download Tools stuff
    def tools_schd_update_downloads(self):
        if self.tool_update_downloads_task is not None:
            return

        self.tool_update_downloads_task = do_update_tools_downloads(
            self.tools_update_downloads, self.tools_update_downloads_ended)

    def tools_update_downloads(self):
        self.header_download.clear()
        tools = self.tool_update_downloads_task.tools_list

        if tools is not None:
            for t in tools:
                itm = TreeItemTool_Download(self.header_download, t)
                self.header_download.addChild(itm)
                itm.updateInTree()

        self.tools_update_downloads_ended()

    def tools_update_downloads_ended(self):
        self.tool_update_downloads_task = None

    ###############################################################################################

    ###############################################################################################
    # Unmanaged Tools stuff
    def tools_schd_update_unmanaged(self):
        if self.tool_update_unmanaged_task is not None:
            return

        self.tool_update_unmanaged_task = do_update_tools_unmanaged(
            self.tools_update_unmanaged, self.tools_update_unmanaged_ended)

    def tools_update_unmanaged(self):
        self.header_unmanaged.clear()
        tools = self.tool_update_unmanaged_task.tools_list
        tools.sort(key=lambda ed: ed.version.as_int, reverse=True)
        self.unmanaged_editors_list = tools

        if tools is not None:
            for t in tools:
                itm = TreeItemTool_ExternallyInstalled(self.header_unmanaged, t)
                self.header_unmanaged.addChild(itm)
                itm.updateInTree()

        self.tools_update_unmanaged_ended()

    def tools_update_unmanaged_ended(self):
        self.tool_update_unmanaged_task = None
        self.header_unmanaged.sortChildren(0, QtCore.Qt.SortOrder.DescendingOrder)

    ###############################################################################################

    ###############################################################################################
    # Managed Tools stuff
    def tools_schd_update_managed(self):
        if self.tool_update_managed_task is not None:
            return

        self.tool_update_managed_task = do_update_tools_managed(
            self.tools_update_managed, self.tools_update_managed_ended)

    def tools_update_managed(self):
        self.header_managed.clear()
        tools = self.tool_update_managed_task.tools_list
        tools.sort(key=lambda ed: ed.version.as_int, reverse=True)
        self.managed_editors_list = tools

        if tools is not None:
            for t in tools:
                itm = TreeItemTool_Managed(self.header_managed, t)
                self.header_managed.addChild(itm)
                itm.updateInTree()

        self.tools_update_managed_ended()

    def tools_update_managed_ended(self):
        self.tool_update_managed_task = None
        self.header_managed.sortChildren(0, QtCore.Qt.SortOrder.DescendingOrder)

    ###############################################################################################

    def open_project_tool(self, game_project: GameProject):
        project_version: Version = game_project.ags_editor_version

        for editor in self.managed_editors_list:
            if editor.version.as_int == project_version.as_int:
                ags_editor_load(editor, game_project)
                return

        for editor in self.unmanaged_editors_list:
            if editor.version.as_int == project_version.as_int:
                ags_editor_load(editor, game_project)
                return

    def build_project_tool(self, game_project: GameProject):
        project_version: Version = game_project.ags_editor_version

        for editor in self.managed_editors_list:
            if editor.version.as_int == project_version.as_int:
                ags_editor_build(editor, game_project)
                return

        for editor in self.unmanaged_editors_list:
            if editor.version.as_int == project_version.as_int:
                ags_editor_build(editor, game_project)
                return

    def filter(self, query: str):
        query = query.lower()
        non_empty_query: bool = query is not None and len(query) != 0
        root = self.invisibleRootItem()
        child_count = root.childCount()
        for i in range(child_count):
            item_header: TreeItemTool_Header = cast(TreeItemTool_Header, root.child(i))
            i_c_count = item_header.childCount()

            for k in range(i_c_count):
                item: TreeItemTool = cast(TreeItemTool, item_header.child(k))
                hidden: bool = False

                if non_empty_query:
                    if item.tool_type == ToolType.MANAGED_TOOL:
                        item_m: TreeItemTool_Managed = cast(TreeItemTool_Managed, item)
                        version_str: str = item_m.itm_wdgt.ags_editor.version.as_str
                        path: str = item_m.itm_wdgt.ags_editor.path
                        path = path.lower()
                        hidden = not(query in version_str or query in path)

                    elif item.tool_type == ToolType.EXTERNALLY_INSTALLED_TOOL:
                        item_e: TreeItemTool_ExternallyInstalled = cast(
                            TreeItemTool_ExternallyInstalled, item)
                        version_str: str = item_e.itm_wdgt.ags_editor.version.as_str
                        path: str = item_e.itm_wdgt.ags_editor.path
                        path = path.lower()
                        hidden = not(query in version_str or query in path)

                    elif item.tool_type == ToolType.AVAILABLE_TO_DOWNLOAD:
                        item_d: TreeItemTool_Download = cast(TreeItemTool_Download, item)
                        version_str: str = item_d.itm_wdgt.release.version.as_str
                        hidden = not(query in version_str)

                item.setHidden(hidden)




from __future__ import annotations # for python 3.8
from typing import Callable, Dict, DefaultDict
from PyQt6 import QtWidgets, QtGui, QtCore


class ExpandableSearchBar(QtWidgets.QLineEdit):
    topWidget: QtWidgets.QWidget = None
    unfocused_width = None
    original_font = None
    focus_functions: Dict[str, Callable[[], None]] = DefaultDict[str, Callable[[], None]]()
    context: Dict[int, str] = DefaultDict[int, str]()
    previous_tab_id: int = None
    searchChanged: QtCore.pyqtSignal = QtCore.pyqtSignal(str)

    def __init__(self, parent: QtWidgets.QWidget = None,
                 top_widget: QtWidgets.QWidget = None):
        super(ExpandableSearchBar, self).__init__(parent)

        self.readyToEdit = None
        self.original_font = self.font()
        self.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding,
                                                 QtWidgets.QSizePolicy.Policy.Fixed))

        self.setClearButtonEnabled(False)
        self.setPlaceholderText("🔍")
        self.setContentsMargins(1, 1, 1, 1)
        self.topWidget = top_widget
        self.unfocused_width = 26
        self.setMaximumWidth(self.unfocused_width)
        self.setMaximumHeight(self.height() - 8)
        self.setMinimumHeight(self.height() - 8)
        self.set_icon_font_if_needed(self.text())
        self.setToolTip("Ctrl+S to select. Type to filter projects or tools")

        shortcut = QtGui.QShortcut(QtGui.QKeySequence("Ctrl+S"), self)
        shortcut.activated.connect(self.setFocus)
        self.textChanged.connect(self.searchChangedEmit)

    def initContext(self, init_ctx_id: int, id_count: int):
        for i in range(id_count):
            self.context[i] = ""
        self.previous_tab_id = init_ctx_id

    def searchChangedEmit(self, text):
        if self.hasFocus() and self.font() != self.original_font:
            self.setFont(self.original_font)
        self.searchChanged.emit(text)

    def context_changed(self, new_context, previous_context):
        if previous_context is not None:
            self.context[previous_context] = self.text()
        if new_context in self.context:
            self.setText(self.context[new_context])
        else:
            self.setText("")
        self.clearFocus()

    def parent_tab_changed(self, index: int):
        self.context_changed(index, self.previous_tab_id)
        self.previous_tab_id = index
        self.set_icon_font_if_needed(self.text())

    def set_parent_focus_functions(self,
                                   f_focus_in: Callable[[QtWidgets.QWidget], None],
                                   f_focus_out: Callable[[QtWidgets.QWidget], None]):

        self.focus_functions['in'] = f_focus_in
        self.focus_functions['out'] = f_focus_out

    def set_icon_font_if_needed(self, txt: str):
        if txt is None or len(txt) == 0:
            fnt = QtGui.QFont("Beedii", 10)
            self.setFont(fnt)
        else:
            self.setFont(self.original_font)

    def tab_parent_changed(self):
        pass

    def mousePressEvent(self, e, Parent=None):
        super(ExpandableSearchBar, self).mousePressEvent(e)
        self.selectAll()

    def focusInEvent(self, event):
        # do custom stuff
        self.selectAll()
        self.setFont(self.original_font)
        self.setPlaceholderText("Search...")
        self.setMaximumWidth(90)
        self.focus_functions['in']()
        super(ExpandableSearchBar, self).focusInEvent(event)
        self.selectAll()

    def focusOutEvent(self, event):
        self.set_icon_font_if_needed(self.text())
        self.setPlaceholderText("🔍")
        self.setMaximumWidth(self.unfocused_width)
        self.focus_functions['out']()
        super(ExpandableSearchBar, self).focusOutEvent(event)



from __future__ import annotations  # for python 3.8
import os
from operator import attrgetter

from agstoolbox.core.ags.ags_editor import EDITOR_FILE_NAME, LocalAgsEditor
from agstoolbox.core.ags.validate_ags_editor import validate_editor_exe
from agstoolbox.core.utils.file import get_gp_candidates_in_dir
from agstoolbox.core.utils.pe import is_valid_exe, get_exe_information
from agstoolbox.core.version.version_utils import version_str_to_version


def is_possibly_editor_file(filepath: str) -> bool:
    if not os.path.exists(filepath):
        return False

    if not filepath.endswith(EDITOR_FILE_NAME):
        return False

    if os.path.getsize(filepath) < 32768 or os.path.getsize(filepath) > 33554432:
        return False

    if not is_valid_exe(filepath):
        return False

    e = get_exe_information(filepath)

    if e.original_filename != 'AGSEditor.exe' or e.internal_name != 'AGSEditor.exe':
        return False

    if e.product_name != 'Adventure Game Studio':
        return False

    return True


def list_probable_ags_editors_paths_in_dir(filepath: str) -> list[str]:
    candidates = get_gp_candidates_in_dir(filepath, EDITOR_FILE_NAME)
    probable_candidates = []
    for c in candidates:
        if is_possibly_editor_file(c):
            probable_candidates.append(c)

    return probable_candidates


def list_probable_ags_editors_in_dir(filepath: str) -> list[LocalAgsEditor]:
    candidates = list_probable_ags_editors_paths_in_dir(filepath)

    editors = []
    for c in candidates:
        local_ae = LocalAgsEditor()
        pe_info = get_exe_information(c)
        version = version_str_to_version(pe_info.product_version)
        local_ae.version = version
        local_ae.path = c
        # TODO: skip validation until we can better maintain validation data
        # local_ae.validated = validate_editor_exe(c, version.as_str)
        local_ae.name = 'AGS Editor ' + version.as_str
        local_ae.last_modified = os.path.getmtime(filepath)
        editors.append(local_ae)

    unique = list(dict.fromkeys(editors))
    editors = unique
    editors.sort(key=attrgetter("last_modified"), reverse=True)

    return editors


def list_ags_editors_in_dir_list(filepaths: list[str]) -> list[LocalAgsEditor]:
    editors = []
    for path in filepaths:
        candidates = list_probable_ags_editors_in_dir(path)
        editors.extend(candidates)

    unique = list(dict.fromkeys(editors))
    editors = unique
    editors.sort(key=attrgetter("last_modified"), reverse=True)

    return editors



import os
import hashlib


from agstoolbox.core.ags.ags_editor_validation_data import \
    AGS_EDITOR_VALIDATED_DATA_ZIP, AGS_EDITOR_VALIDATED_DATA_CONTENTS

BUF_SIZE = 131072


def validate_file(filepath, filename, validation_data):
    if filename not in validation_data:
        return False

    if not os.path.exists(filepath):
        return False

    if os.path.isdir(filepath):
        return False

    if os.path.getsize(filepath) != validation_data[filename]['size']:
        return False

    md5_checker = hashlib.md5()
    with open(filepath, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            md5_checker.update(data)

    return md5_checker.hexdigest() == validation_data[filename]['md5']


def validate_ags_editor_zip(z_file):
    return validate_file(z_file, os.path.basename(z_file), AGS_EDITOR_VALIDATED_DATA_ZIP)


def validate_editor_contents(filepath, filename, version):
    return validate_file(filepath, filename, AGS_EDITOR_VALIDATED_DATA_CONTENTS[version])


def validate_editor_exe(filepath, version):
    if version not in AGS_EDITOR_VALIDATED_DATA_CONTENTS.keys():
        return False
    return validate_file(filepath, os.path.basename(filepath),
                         AGS_EDITOR_VALIDATED_DATA_CONTENTS[version])



from __future__ import annotations  # for python 3.8
from pathlib import Path

from PyQt6 import QtCore
from PyQt6.QtWidgets import QWidget, QFileDialog, QLineEdit, QPushButton, QHBoxLayout

from agstoolbox.core.settings.settings import ConstSettings
from agstoolbox.core.utils.file import dir_is_valid


class DirEditWidget(QWidget):
    default_dir_value = None

    def __init__(self, initial_dir: str, default_dir: str, parent: QWidget = None):
        QWidget.__init__(self, parent)

        self.dir_line_edit = QLineEdit(self)
        self.dir_line_edit.setEnabled(False)
        self.dir_line_edit.setObjectName("dir_line_edit")

        self.dir_edit_button = QPushButton(self)
        self.dir_edit_button.setObjectName("dir_edit_button")

        self.dir_defaults_button = QPushButton(self)
        self.dir_defaults_button.setObjectName("dir_defaults_button")

        self.h_layout = QHBoxLayout(self)
        self.h_layout.setObjectName("h_layout")
        self.h_layout.addWidget(self.dir_line_edit)
        self.h_layout.addWidget(self.dir_edit_button)
        self.h_layout.addWidget(self.dir_defaults_button)
        self.setLayout(self.h_layout)

        self.dir_line_edit.setText(initial_dir)
        self.default_dir_value = default_dir

        self.retranslateUi()

        self.dir_edit_button.clicked.connect(self.btn_edit_clicked)
        self.dir_defaults_button.clicked.connect(self.btn_defaults_clicked)

    def retranslateUi(self):
        _translate = QtCore.QCoreApplication.translate
        parent = ""
        if self.parent() is not None:
            parent = self.parent().objectName()

        self.dir_edit_button.setText(_translate(parent, "Edit"))
        self.dir_defaults_button.setText(_translate(parent, "Set Defaults"))

    def btn_edit_clicked(self):
        dir_path = self.dir_line_edit.text()
        if not dir_is_valid(dir_path):
            dir_path = ConstSettings().user_docs

        dir_path = QFileDialog.getExistingDirectory(
            self, 'Select Folder',
            options=QFileDialog.Option.ShowDirsOnly,
            directory=dir_path)

        if not dir_is_valid(dir_path):
            return

        dir_path = str(Path(str(dir_path)).as_posix())
        self.dir_line_edit.setText(dir_path)

    def btn_defaults_clicked(self):
        self.dir_line_edit.setText(self.default_dir_value)

    def setText(self, text: str):
        self.dir_line_edit.setText(text)

    def text(self) -> str:
        return self.dir_line_edit.text()



from __future__ import annotations  # for python 3.8
from agstoolbox.core.version.version import Version

PROJECT_FILE_NAME = 'Game.agf'


class GameProject:
    """an AGS Game Project"""
    name: str | None = None
    game_file: str | None = None
    ags_editor_version: Version = None
    ags_editor_version_index = None
    path: str | None = None
    directory: str | None = None
    encoding: str | None = None
    codepage: int | None = None
    last_modified = None
    ico_path = None



from __future__ import annotations  # for python 3.8
import sys

# progress bar logic extracted from 'progress' module from Georgios Verigakis
# we are using a very simplified version to work in most terminals (in win, and linux, ...)


class ProgressBar:
    if sys.platform.startswith('win'):
        phases = (' ', '▌', '█')
    else:
        phases = (' ', '▏', '▎', '▍', '▌', '▋', '▊', '▉', '█')

    def __init__(self, l_text: str, r_text: str, max_value: int, unit: str = '', width: int = 32):
        self.max: int = max_value
        self.width: int = width
        self.progress: float = 0.0
        self.l_text: str = l_text
        self.r_text: str = r_text
        self.unit = unit

    def update(self, current_value: int, total_size: int):
        self.max = total_size
        self.progress: float = current_value / self.max
        filled_len = self.width * self.progress
        nfull: int = int(filled_len)
        phase: int = int((filled_len - nfull) * len(self.phases))
        nempty: int = self.width - nfull

        bar: str = self.phases[-1] * nfull
        current: str = self.phases[phase] if phase > 0 else ''
        empty: str = ' ' * max(0, nempty - len(current))
        line = f"{current_value}/{self.max} {self.unit} |{bar}{current}{empty}|"

        sys.stdout.write('\r' + self.l_text + line + self.r_text)
        sys.stdout.flush()

    def finish(self):
        self.progress = self.max
        sys.stdout.write('\n')
        sys.stdout.flush()



from __future__ import annotations  # for python 3.8

import os
import sys
from subprocess import Popen, PIPE, TimeoutExpired

from agstoolbox.core.utils.file import get_file, get_dir
from agstoolbox.core.utils.pyinstaller_hacks import lock_dll_dir, unlock_dll_dir


def run_exe_params(exe_path: str, block: bool = False, params=None) -> int:
    if params is None:
        params = []
    exe_file = get_file(exe_path)
    working_dir = get_dir(exe_path)

    p_params = [exe_file]
    p_params.extend(params)

    cwd = os.getcwd()
    os.chdir(working_dir)
    lock_dll_dir()
    print('Popen: cwd=' + working_dir + ', ' + ' '.join(p_params))
    proc = Popen(p_params, cwd=working_dir, stdout=PIPE, stderr=PIPE, bufsize=1,
                 universal_newlines=True)
    count = 0
    if block:
        while count < 1000:
            for line in proc.stdout:
                sys.stdout.write(line)
            for line in proc.stderr:
                sys.stdout.write(line)
            try:
                proc.wait(10)
            except TimeoutExpired:
                pass
            count += 1

        if count == 1000:
            proc.terminate()
    return_code: int = proc.returncode
    unlock_dll_dir()
    os.chdir(cwd)
    return return_code



from __future__ import annotations  # for python 3.8
import os

from PyQt6 import QtGui, QtCore


import subprocess
import sys


class UnsupportedPlatformException(Exception):
    pass


def _show_file_darwin(path):
    subprocess.check_call(["open", "--", path])


def _show_file_linux(path):
    subprocess.check_call(["xdg-open", "--", path])


def _show_file_win32(path):
    subprocess.check_call(["explorer", "/select", path])


def open_folder(path: str):
    _show_file_func = {'darwin': _show_file_darwin,
                       'linux': _show_file_linux,
                       'win32': _show_file_win32}
    fullpath = os.path.realpath(path)
    if not QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(fullpath)):
        # failed
        try:
            show_file = _show_file_func[sys.platform](fullpath)
        except KeyError:
            raise UnsupportedPlatformException



from __future__ import annotations

import sys


def _win32_unsafe_remove_app_at_startup(app_name: str):
    from agstoolbox.core.utils.win_registry import set_run_key
    set_run_key(app_name, None)


def _win32_remove_app_at_startup(app_name: str):
    try:
        _win32_unsafe_remove_app_at_startup(app_name)
    except OSError:
        print("could not remove app from startup")


def _win32_set_app_at_startup(app_name: str, app_path: str) -> bool:
    from agstoolbox.core.utils.win_registry import set_run_key
    set_failed = False

    try:
        set_run_key(app_name, app_path)
    except OSError:
        set_failed = True

    if set_failed:
        try:
            _win32_unsafe_remove_app_at_startup(app_name)
        except OSError:
            set_failed = True

    return not set_failed


def _linux_remove_app_at_startup(app_name: str):
    print("stub _linux_remove_app_at_startup")


def _linux_set_app_at_startup(app_name: str, app_path: str) -> bool:
    print("stub _linux_set_app_at_startup")
    return False


def _darwin_remove_app_at_startup(app_name: str):
    print("stub _darwin_remove_app_at_startup")


def _darwin_set_app_at_startup(app_name: str, app_path: str) -> bool:
    print("stub _darwin_set_app_at_startup")
    return False


def remove_app_at_startup(app_name: str):
    _remove_app_at_startup = {
        'darwin': _darwin_remove_app_at_startup,
        'linux': _linux_remove_app_at_startup,
        'win32': _win32_remove_app_at_startup}

    _remove_app_at_startup[sys.platform](app_name)


def set_app_at_startup(app_name: str, app_path: str) -> bool:
    _set_app_at_startup = {
        'darwin': _darwin_set_app_at_startup,
        'linux': _linux_set_app_at_startup,
        'win32': _win32_set_app_at_startup}

    return _set_app_at_startup[sys.platform](app_name, app_path)



import os
import zipfile

from agstoolbox.core.gh.download_release import get_zip_archive_cache_path
from agstoolbox.core.gh.release import Release
from agstoolbox.core.settings.settings import Settings
from agstoolbox.core.utils.file import remove_dir_contents, mkdirp


def get_release_install_dir(release: Release) -> str:
    base_install_dir = Settings().get_editor_install_dir()
    return os.path.join(base_install_dir, release.version.as_str)


def is_install_dir_busy(release: Release) -> bool:
    install_dir = get_release_install_dir(release)
    dir_doesnt_exist: bool = not os.path.exists(install_dir)
    if dir_doesnt_exist:
        return False
    dir_exists_but_empty: bool = os.path.isdir(install_dir) and not os.listdir(install_dir)
    return not dir_exists_but_empty


def install_release_from_cache(release: Release):
    filepath_in_cache = get_zip_archive_cache_path(release)
    install_dir = get_release_install_dir(release)
    remove_dir_contents(install_dir)
    mkdirp(install_dir)
    with zipfile.ZipFile(filepath_in_cache, 'r') as zip_ref:
        zip_ref.extractall(install_dir)



from __future__ import annotations  # for python 3.8

import os.path
from struct import pack
from typing import BinaryIO

from agstoolbox.core.ags.script_module import (ScriptModule, MODULE_FILE_SIGNATURE,
                                               MODULE_FILE_SECTION, MODULE_FILE_TRAILER)
from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.ags.get_script_module import module_from_game_project


def export_script_module_from_project(game_project: GameProject, module_name: str, out: str | None):
    if out is None:
        out = game_project.directory

    sm: ScriptModule = module_from_game_project(game_project, module_name)
    export_script_module(sm, out, game_project.encoding, game_project.codepage)


def export_script_module(sm: ScriptModule, out_dir: str, enc: str, codepage: int):
    module_filename = os.path.join(out_dir, sm.basename + ".scm")

    with open(module_filename, 'wb') as f:
        f.write(MODULE_FILE_SIGNATURE)
        f.write(pack('i', 1))  # version

        write_string_terminated(sm.author, enc, f)
        write_string_terminated(sm.description, enc, f)
        write_string_terminated(sm.name, enc, f)
        write_string_terminated(sm.version, enc, f)

        write_string_long_terminated(sm.script, enc, f)
        write_string_long_terminated(sm.header, enc, f)

        f.write(pack('i', sm.unique_key_int))
        f.write(pack('i', 0))  # Permissions (obsolete)
        f.write(pack('i', 0))  # We are owner (obsolete)

        # format extension 1
        f.write(pack('I', MODULE_FILE_SECTION))
        f.write(pack('I', codepage))

        # end of format
        f.write(pack('I', MODULE_FILE_TRAILER))


def write_string_terminated(text: str, enc: str, f: BinaryIO):
    text_bytes: bytes = text.encode(encoding=enc)
    f.write(text_bytes)
    f.write(b'\x00')


def write_string_long_terminated(text: str, enc: str, f: BinaryIO):
    text_bytes: bytes = text.encode(encoding=enc)
    f.write(pack('I', len(text_bytes)))
    f.write(text_bytes)
    f.write(b'\x00')




from __future__ import annotations

import os
import sys


def is_running_from_pyinstaller() -> bool:
    return getattr(sys, 'frozen', False) is not False


def get_extracted_path() -> str:
    if hasattr(sys, '_MEIPASS'):
        return str(os.path.join(sys._MEIPASS))  # pylint: disable=no-member
    else:
        return str("")


def _win32_requires_dll_locking() -> bool:
    if is_running_from_pyinstaller():
        ext_path: str = get_extracted_path()
        return len(ext_path) > 0
    else:
        return False


def _do_nothing():
    pass


def _win32_lock_dll_dir():
    if _win32_requires_dll_locking():
        import ctypes
        ctypes.windll.kernel32.SetDllDirectoryW(None)


def _win32_unlock_dll_dir():
    if _win32_requires_dll_locking():
        import ctypes
        ctypes.windll.kernel32.SetDllDirectoryW(get_extracted_path())


def lock_dll_dir():
    _lock_dll_dir = {
        'darwin': _do_nothing,
        'linux': _do_nothing,
        'win32': _win32_lock_dll_dir}

    _lock_dll_dir[sys.platform]()


def unlock_dll_dir():
    _unlock_dll_dir = {
        'darwin': _do_nothing,
        'linux': _do_nothing,
        'win32': _win32_unlock_dll_dir}

    _unlock_dll_dir[sys.platform]()



from __future__ import annotations  # for python 3.8
from agstoolbox.core.version.version import Version

EDITOR_FILE_NAME = 'AGSEditor.exe'


class AgsEditor:
    version: Version = None
    name = None


class LocalAgsEditor(AgsEditor):
    path = None
    externally_installed: bool = False
    validated = None
    last_modified = None



from __future__ import annotations  # for python 3.8

from PyQt6.QtWidgets import QApplication, QWidget
from PyQt6 import QtGui

from agstoolbox import getdata
from agstoolbox.core.utils.singleton import Singleton


class StaticFontData:
    custom_fonts = {}
    parent: QApplication = None


class FontData(StaticFontData, metaclass=Singleton):
    pass


def add_font(font_id: str, filepath: str) -> None:
    if FontData.custom_fonts is None:
        FontData.custom_fonts = {}

    FontData.custom_fonts[font_id] = QtGui.QFontDatabase.addApplicationFont(getdata.path(filepath))


def set_custom_fonts(app: QApplication, main_widget: QWidget) -> None:
    FontData.parent = app
    FontData.custom_fonts = {}

    add_font('Beedii', "fonts/Beedii.ttf")


def get_custom_font_id(font_id: str) -> int:
    return FontData.custom_fonts[font_id]



from __future__ import annotations  # for python 3.8

import os.path

import defusedxml.ElementTree as ETree
from xml.etree.ElementTree import Element

from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.ags.script_module import ScriptModule


def _module_element_from_project(game_project: GameProject, module_name: str) -> Element | None:
    tree = ETree.parse(game_project.path)
    root = tree.getroot()
    scripts_and_header_htags = root.findall('.//ScriptAndHeader')
    script_htag: Element | None = None
    for sah_htag in scripts_and_header_htags:
        sah_script_htag = sah_htag.find('ScriptAndHeader_Script')
        script_htag = sah_script_htag.find('Script')
        script_name_htag = script_htag.find('Name')
        if script_name_htag.text == module_name:
            break
        script_fname_htag = script_htag.find('FileName')
        if script_fname_htag.text[:-4] == module_name:
            break
        script_htag = None

    if script_htag is None:
        return None
    return script_htag


def exists_module_in_game_project(game_project: GameProject, module_name: str) -> bool:
    script_htag = _module_element_from_project(game_project, module_name)
    return (script_htag is not None) and True


def module_from_game_project(game_project: GameProject, module_name: str) -> ScriptModule | None:
    script_htag = _module_element_from_project(game_project, module_name)

    if script_htag is None:
        return None

    script_filename_htag = script_htag.find('FileName')
    module_filename = script_filename_htag.text[:-4]
    header_filename = module_filename + ".ash"
    script_filename = module_filename + ".asc"
    full_header_filename = os.path.abspath(os.path.join(game_project.directory, header_filename))
    full_script_filename = os.path.abspath(os.path.join(game_project.directory, script_filename))

    sm: ScriptModule = ScriptModule()
    sm.basename = module_filename
    sm.name = script_htag.find('Name').text
    sm.version = script_htag.find('Version').text
    sm.author = script_htag.find('Author').text
    sm.unique_key = script_htag.find('Key').text
    sm.description = script_htag.find('Description').text

    sm.unique_key_int = int(sm.unique_key)

    with open(full_header_filename, "r") as f:
        sm.header = f.read()

    with open(full_script_filename, "r") as f:
        sm.script = f.read()

    return sm



from __future__ import annotations  # for python 3.8


MODULE_FILE_SIGNATURE = b'AGSScriptModule\0'
MODULE_FILE_SECTION = 0xb4f76a65
MODULE_FILE_TRAILER = 0xb4f76a66


class ScriptModule:
    basename: str | None = None
    name: str | None = None
    version: str | None = None
    author: str | None = None
    description: str | None = None
    unique_key: str | None = None
    unique_key_int: int | None = None
    script: str | None = None
    header: str | None = None



from __future__ import annotations

from PyQt6 import QtCore, QtWidgets, QtNetwork

from agstoolbox.system.at_runguard import RunGuard


class unique_application(QtWidgets.QApplication):
    _run_guard: RunGuard = None
    _socket: QtNetwork.QLocalSocket = None
    _key: str = None
    _listener: QtNetwork.QLocalServer = None

    another_instance = QtCore.pyqtSignal()
    is_unique: bool = False

    def __init__(self, title: str, argv: list[str]):
        QtWidgets.QApplication.__init__(self, argv)
        self._key = "18d1411b-54a1-46fc-bd14-1d38519e2c61" + title
        self.setApplicationName(title)
        self._run_guard = RunGuard(parent=self, key=self._key)

        if not self._run_guard.unique:
            self.is_unique = False
            _socket = QtNetwork.QLocalSocket()
            _socket.connectToServer(self._key)
            _socket.close()
            return
        else:
            self.is_unique = True
            self.listener = QtNetwork.QLocalServer(self)
            self.listener.setSocketOptions(QtNetwork.QLocalServer.SocketOption.WorldAccessOption)
            self.listener.newConnection.connect(self.another_instance)
            self.listener.listen(self._key)




from __future__ import annotations  # for python 3.8
from typing import Callable

import requests


def download_from_url(url: str, save_path: str, chunk_size: int = 4096,
                      progress_update: Callable[[float, int, int, int], None] = None):
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get('content-length', 0))
    acc_c_size = 0
    with open(save_path, 'wb') as fd:
        for chunk in response.iter_content(chunk_size=chunk_size):
            c_size = len(chunk)*8
            acc_c_size += c_size
            completed_percent = acc_c_size / total_size_in_bytes
            if progress_update is not None:
                progress_update(completed_percent, c_size, acc_c_size, total_size_in_bytes)
            fd.write(chunk)



from PyQt6 import QtCore, QtWidgets
from PyQt6.QtWidgets import QDialog, QSizePolicy, QFormLayout, QHBoxLayout, QVBoxLayout, \
    QSpacerItem, QLabel, QDialogButtonBox

from agstoolbox.core.settings.settings import Settings, ConstSettings
from agstoolbox.wdgts.at_dirlist_wdgt import DirListWidget
from agstoolbox.wdgts.at_single_dir_wdgt import DirEditWidget
from agstoolbox.wdgts_utils.get_self_path import get_app_path
from agstoolbox import __version__


class SettingsDialog(QDialog):
    def __init__(self, parent: QtWidgets = None):
        QDialog.__init__(self, parent)
        self.setObjectName("SettingsDialog")
        self.resize(512, 400)
        self.setSizeGripEnabled(True)

        self.label_settings_intro = QLabel(self)
        self.label_settings_intro.setObjectName("label_settings_intro")
        self.label_settings_intro.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self.run_at_startup_label = QLabel(self)
        self.run_at_startup_label.setObjectName("run_at_startup_label")
        self.run_at_startup_checkbox = QtWidgets.QCheckBox(self)
        self.run_at_startup_checkbox.setObjectName("run_at_startup_checkbox")

        self.base_install_dir_label = QLabel(self)
        self.base_install_dir_label.setObjectName("base_install_dir_label")

        self.install_dir_line_edit = DirEditWidget(
            parent=self,
            initial_dir=Settings().get_tools_install_dir(),
            default_dir=ConstSettings().DEFAULT_TOOLS_INSTALL_DIR)

        self.label_editors = QtWidgets.QLabel(self)
        self.label_editors.setWordWrap(True)
        self.label_editors.setObjectName("label_editors")

        self.external_editors_dir_search_list = DirListWidget(
            parent=self,
            default_dirs=ConstSettings().DEFAULT_EXT_EDITORS_SEARCH_DIRS,
            dirs=[])
        self.external_editors_dir_search_list.setObjectName("external_editors_dir_search_list")

        self.label_projects = QtWidgets.QLabel(self)
        self.label_projects.setWordWrap(True)
        self.label_projects.setObjectName("label_projects")

        self.project_dir_search_list = DirListWidget(
            parent=self,
            default_dirs=ConstSettings().DEFAULT_PROJECTS_SEARCH_DIRS,
            dirs=[])
        self.project_dir_search_list.setObjectName("project_dir_search_list")

        self.button_box = QDialogButtonBox(self)
        self.button_box.setStandardButtons(
            QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        self.button_box.setObjectName("buttonBox")

        self.formLayout = QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.horizontalLayout_3 = QHBoxLayout(self)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.verticalLayout_3 = QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.verticalLayout_3.addWidget(self.label_settings_intro)

        # run at startup
        self.formLayout.setWidget(1, QFormLayout.ItemRole.FieldRole,
                                  self.run_at_startup_checkbox)
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.run_at_startup_label)

        # manual editor search dirs
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole,
                                  self.external_editors_dir_search_list)
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.label_editors)

        # project search dirs
        self.formLayout.setWidget(3, QFormLayout.ItemRole.FieldRole,
                                  self.project_dir_search_list)
        self.formLayout.setWidget(3, QFormLayout.ItemRole.LabelRole, self.label_projects)

        # install tools dir
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.ItemRole.FieldRole,
                                  self.install_dir_line_edit)
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.ItemRole.LabelRole,
                                  self.base_install_dir_label)

        self.verticalLayout_3.addLayout(self.formLayout)
        spacer_item = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum,
                                  QSizePolicy.Policy.Expanding)
        self.verticalLayout_3.addItem(spacer_item)
        self.verticalLayout_3.addWidget(self.button_box)
        self.horizontalLayout_3.addLayout(self.verticalLayout_3)

        self.retranslateUi()
        QtCore.QMetaObject.connectSlotsByName(self)

        self.apply_from_settings_to_dialog()
        self.button_box.accepted.connect(self.clicked_ok)
        self.button_box.rejected.connect(self.clicked_cancel)

    def apply_from_settings_to_dialog(self):
        Settings().set_app_path(get_app_path())

        run_at_startup = Settings().get_run_when_os_starts()
        self.run_at_startup_checkbox.setChecked(run_at_startup)

        dirs = Settings().get_manually_installed_editors_search_dirs()
        self.external_editors_dir_search_list.setDirs(dirs)

        dirs = Settings().get_project_search_dirs()
        self.project_dir_search_list.setDirs(dirs)

        install_dir = Settings().get_tools_install_dir()
        self.install_dir_line_edit.setText(install_dir)

    def apply_from_dialog_to_settings(self):
        Settings().set_app_path(get_app_path())

        run_at_startup = self.run_at_startup_checkbox.isChecked()
        Settings().set_run_when_os_starts(run_at_startup)

        dirs = self.external_editors_dir_search_list.getDirs()
        Settings().set_manually_installed_editors_search_dirs(dirs)

        dirs = self.project_dir_search_list.getDirs()
        Settings().set_project_search_dirs(dirs)

        install_dir = self.install_dir_line_edit.text()
        try:
            Settings().set_tools_install_dir(install_dir)
        except ValueError:
            QtWidgets.QMessageBox.warning(
                self, "Warning", "tools installation dir not found and not set.")

        try:
            Settings().save()
        except OSError:
            QtWidgets.QMessageBox.warning(
                self, "Warning", "failed to save settings.")

    def retranslateUi(self):
        _translate = QtCore.QCoreApplication.translate
        self.setWindowTitle(_translate("SettingsDialog", "Settings"))
        self.label_settings_intro.setText(
            "AGS Toolbox " + __version__ + ". " +
            _translate("SettingsDialog", "Adjust settings here."))
        self.run_at_startup_label.setText(
            _translate("SettingsDialog",
                       "Run on OS startup (experimental)"))
        self.base_install_dir_label.setText(_translate("SettingsDialog", "Base install dir"))
        self.label_editors.setText(
            _translate("SettingsDialog", "Externally installed AGS Editors search paths"))
        self.label_projects.setText(
            _translate("SettingsDialog", "AGS Game projects search paths"))
        self.install_dir_line_edit.retranslateUi()
        self.project_dir_search_list.retranslateUi()
        self.external_editors_dir_search_list.retranslateUi()

    def closeEvent(self, evnt):
        QDialog.closeEvent(self, evnt)

    def clicked_ok(self):
        self.accept()

    def clicked_cancel(self):
        self.reject()

    def accept(self) -> None:
        self.apply_from_dialog_to_settings()
        self.apply_from_settings_to_dialog()
        QDialog.accept(self)

    def reject(self) -> None:
        self.apply_from_settings_to_dialog()
        QDialog.reject(self)



from __future__ import annotations  # for python 3.8
from enum import Enum
from PyQt6 import QtCore, QtGui, QtWidgets
from PyQt6.QtWidgets import QTreeWidgetItem, QLabel, QHBoxLayout, QVBoxLayout, QGridLayout, QWidget

from agstoolbox.at_icons import ags_editor_as_pixmap
from agstoolbox.at_tasks import do_download_managed
from agstoolbox.core.ags.ags_editor import LocalAgsEditor
from agstoolbox.core.ags.ags_local_run import ags_editor_start
from agstoolbox.core.utils.open_in_browser import open_in_browser
from agstoolbox.wdgts_utils.ags_local_extra import ags_editor_folder_in_explorer
from agstoolbox.core.gh.release import Release
from agstoolbox.core.utils.time import s_ago
from agstoolbox.wdgts_utils.action_utils import DefaultMenuQAction


class ToolType(Enum):
    HEADER = 16
    MANAGED_TOOL = 1
    AVAILABLE_TO_DOWNLOAD = 2
    EXTERNALLY_INSTALLED_TOOL = 4

    MANAGED_TOOL_HEADER = 17
    AVAILABLE_TO_DOWNLOAD_HEADER = 18
    EXTERNALLY_INSTALLED_TOOL_HEADER = 20


class TreeItemTool(QTreeWidgetItem):
    tool_type: ToolType = None

    def __init__(self, parent: QTreeWidgetItem | None, tool_type: ToolType):
        QTreeWidgetItem.__init__(self, parent)
        self.tool_type = tool_type


class EditorImage(QLabel):
    def __init__(self, parent: QWidget = None):
        QLabel.__init__(self, parent)

        pxmap = ags_editor_as_pixmap()

        self.setPixmap(pxmap.scaled(
            32, 32,
            QtCore.Qt.AspectRatioMode.KeepAspectRatio,
            QtCore.Qt.TransformationMode.SmoothTransformation))
        self.setMaximumSize(32, 32)
        self.setMinimumSize(32, 32)


class TreeItemTool_Header(TreeItemTool):
    name = None

    def __init__(self, name: str, tool_type: ToolType):
        TreeItemTool.__init__(self, None, tool_type)
        self.name = name
        self.setWhatsThis(0, name)
        self.tool_type = tool_type
        self.setText(0, name)

    def __lt__(self, other):
        return False  # keeps unaltered when sorting

    def clear(self):
        for i in range(self.childCount()):
            self.removeChild(self.child(0))


# Tool Download Item
class TreeItemTool_Download_Widget(QWidget):
    release = None
    # we need to store the thread or the garbage collector will remove it!
    thread_download = None

    def __init__(self, release: Release, parent: QWidget = None):
        QWidget.__init__(self, parent)
        self.release = release

        self.icon_img = EditorImage()
        self.labelName = QLabel(self.release.name)
        self.labelName.setWordWrap(True)

        smaller_font = QtGui.QFont(
            self.labelName.font().family(),
            self.labelName.font().pointSize() * 0.90,
        )

        smallest_font = QtGui.QFont(
            self.labelName.font().family(),
            self.labelName.font().pointSize() * 0.75,
        )

        self.labelTime = QLabel(s_ago(self.release.published_at_timestamp))
        self.labelTime.setFont(smaller_font)
        self.labelTime.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self.labelDir = QLabel(self.release.url)
        self.labelDir.setFont(smallest_font)
        self.labelDir.setDisabled(True)

        top_hbox = QHBoxLayout()
        bottom_hbox = QHBoxLayout()

        right_vbox = QVBoxLayout()

        vbox = QVBoxLayout()

        vbox.addLayout(top_hbox)
        vbox.addLayout(bottom_hbox)
        top_hbox.addWidget(self.labelName)
        right_vbox.addWidget(self.labelTime)
        top_hbox.addLayout(right_vbox)
        bottom_hbox.addWidget(self.labelDir)

        main_qgrid = QGridLayout()
        main_qgrid.addWidget(self.icon_img)
        main_qgrid.addLayout(vbox, 0, 1)

        self.setLayout(main_qgrid)

    def install(self):
        if self.thread_download is not None:
            return
        self.thread_download = do_download_managed(self.release, self.download_ended)
        self.setEnabled(False)

    def show_gh_rel_page(self):
        open_in_browser(self.release.html_url)

    def mouseDoubleClickEvent(self, event):
        self.install()

    def download_ended(self):
        self.thread_download = None
        self.parent().parent().tools_schd_update_managed()
        self.setEnabled(True)

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu(self)
        install_action = DefaultMenuQAction(menu, "Install as managed Editor")
        show_gh_rel_page_action = menu.addAction("Show GitHub Release Page")
        action = menu.exec(self.mapToGlobal(event.pos()))
        if action == install_action:
            self.install()
        elif action == show_gh_rel_page_action:
            self.show_gh_rel_page()


# tool available locally for use
# either managed or unmanaged (externally installed)
class TreeItemTool_Local_Widget(QWidget):
    ags_editor = None

    def __init__(self, ags_editor: LocalAgsEditor, tool_type: ToolType, parent: QWidget = None):
        QWidget.__init__(self, parent)
        self.ags_editor = ags_editor

        self.icon_img = EditorImage()
        self.labelName = QLabel(self.ags_editor.name)
        self.labelName.setWordWrap(True)

        smaller_font = QtGui.QFont(
            self.labelName.font().family(),
            self.labelName.font().pointSize() * 0.90,
        )

        smallest_font = QtGui.QFont(
            self.labelName.font().family(),
            self.labelName.font().pointSize() * 0.75,
        )

        self.labelVersion = QLabel(self.ags_editor.version.as_str)
        self.labelVersion.setFont(smaller_font)
        self.labelVersion.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self.labelTime = QLabel(s_ago(self.ags_editor.last_modified))
        self.labelTime.setFont(smaller_font)
        self.labelTime.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self.labelDir = QLabel(self.ags_editor.path)
        self.labelDir.setFont(smallest_font)
        self.labelDir.setDisabled(True)

        top_hbox = QHBoxLayout()
        bottom_hbox = QHBoxLayout()

        right_vbox = QVBoxLayout()

        vbox = QVBoxLayout()

        vbox.addLayout(top_hbox)
        vbox.addLayout(bottom_hbox)
        top_hbox.addWidget(self.labelName)
        right_vbox.addWidget(self.labelTime)
        right_vbox.addWidget(self.labelVersion)
        top_hbox.addLayout(right_vbox)
        bottom_hbox.addWidget(self.labelDir)

        main_qgrid = QGridLayout()
        main_qgrid.addWidget(self.icon_img)
        main_qgrid.addLayout(vbox, 0, 1)

        self.setLayout(main_qgrid)

    def open_editor(self):
        ags_editor_start(self.ags_editor)

    def mouseDoubleClickEvent(self, event):
        self.open_editor()

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu(self)
        open_editor_action = DefaultMenuQAction(menu, "Open Editor")
        open_folder_action = menu.addAction("Open Folder in File Explorer")
        action = menu.exec(self.mapToGlobal(event.pos()))
        if action == open_editor_action:
            self.open_editor()
        elif action == open_folder_action:
            ags_editor_folder_in_explorer(self.ags_editor)


##################################################################################################
# item widgets

class TreeItemTool_Managed(TreeItemTool):
    def __init__(self, parent: QTreeWidgetItem, ags_editor: LocalAgsEditor):
        TreeItemTool.__init__(self, parent, ToolType.MANAGED_TOOL)
        self.itm_wdgt = TreeItemTool_Local_Widget(ags_editor, self.tool_type)

    def __lt__(self, other):
        return self.itm_wdgt.ags_editor.version.as_int < other.itm_wdgt.ags_editor.version.as_int

    def updateInTree(self):
        self.treeWidget().setItemWidget(self, 0, self.itm_wdgt)


class TreeItemTool_ExternallyInstalled(TreeItemTool):
    def __init__(self, parent: QTreeWidgetItem, ags_editor: LocalAgsEditor):
        TreeItemTool.__init__(self, parent, ToolType.EXTERNALLY_INSTALLED_TOOL)
        self.itm_wdgt = TreeItemTool_Local_Widget(ags_editor, self.tool_type)

    def __lt__(self, other):
        return self.itm_wdgt.ags_editor.version.as_int < other.itm_wdgt.ags_editor.version.as_int

    def updateInTree(self):
        self.treeWidget().setItemWidget(self, 0, self.itm_wdgt)


class TreeItemTool_Download(TreeItemTool):
    def __init__(self, parent: QTreeWidgetItem, release: Release):
        TreeItemTool.__init__(self, parent, ToolType.AVAILABLE_TO_DOWNLOAD)
        self.itm_wdgt = TreeItemTool_Download_Widget(release)

    def updateInTree(self):
        self.treeWidget().setItemWidget(self, 0, self.itm_wdgt)



from __future__ import annotations  # for python 3.8

from PyQt6.QtGui import QAction, QFont
from PyQt6.QtWidgets import QMenu


def DefaultMenuQAction(menu: QMenu, action_text: str) -> QAction:
    action = menu.addAction(action_text)
    font: QFont = action.font()
    font.setBold(True)
    action.setFont(font)
    menu.addSeparator()
    return action



from __future__ import annotations  # for python 3.8
import pefile


class ExeFileDescriptor:
    product_name = None
    original_filename = None
    internal_name = None
    file_version = None
    product_version = None
    company_name = None


def is_valid_exe(filepath: str) -> bool:
    is_exe = False
    try:
        pe = pefile.PE(filepath)
        is_exe = pe.is_exe()
    except pefile.PEFormatError:
        return False
    finally:
        return is_exe


def get_exe_information(filepath: str) -> ExeFileDescriptor:
    string_info = {}
    pe = None
    try:
        pe = pefile.PE(filepath)
    except pefile.PEFormatError:
        return ExeFileDescriptor()

    if not pe.is_exe():
        return ExeFileDescriptor()

    for fileinfo in pe.FileInfo[0]:
        if fileinfo.Key.decode() == 'StringFileInfo':
            for st in fileinfo.StringTable:
                for entry in st.entries.items():
                    string_info[entry[0].decode()] = entry[1].decode()

    bin_exe = ExeFileDescriptor()
    bin_exe.product_name = string_info['ProductName']  # should be 'Adventure Game Studio'
    bin_exe.original_filename = string_info['OriginalFilename']  # should be 'AGSEditor.exe'
    bin_exe.product_version = string_info['ProductVersion']  # should be something like '3.5.1.14'
    bin_exe.file_version = string_info['FileVersion']  # should be something like '3.5.1.14'
    bin_exe.internal_name = string_info['InternalName']  # should be 'AGSEditor.exe'
    bin_exe.company_name = string_info['CompanyName']  # should be 'AGS'

    return bin_exe



from __future__ import annotations

from agstoolbox.core.version.version import Version


def tag_to_version_str(tag: str) -> str:
    if tag is None or "".__eq__(tag):
        return ""

    tks = tag.split(".")
    if len(tks) > 5:
        return tag

    first_tk = 0
    if tks[0][0] == 'v' and len(tks[0]) > 1:
        if tks[0][1] != '.' and tks[0][1].isnumeric():
            tks[0] = tks[0][1:]

    if (tks[0][0] == 'v' and len(tks[0]) <= 2) or tks[0] == "version":
        first_tk = 1

    if not tks[first_tk].isnumeric():
        return tag

    major: str = "0"
    minor: str = "0"
    improv: str = "0"
    patch: str = "0"

    if len(tks) >= 1:
        major = tks[first_tk].strip()

    if len(tks) >= 2:
        minor = tks[first_tk + 1].strip()

    if len(tks) >= 3:
        improv = tks[first_tk + 2].strip()

    if len(tks) >= 4:
        patch = tks[first_tk + 3].strip()

    return major + "." + minor + "." + improv + "." + patch


def tag_to_family(tag: str) -> str:
    tks = tag_to_version_str(tag).split(".")
    if len(tks) <= 1 or len(tks) > 5:
        return tag

    major = tks[0]
    minor = tks[1]

    if not major.isnumeric() and not minor.isnumeric():
        return tag

    family = major + "." + minor
    return family


def family_to_major(family: str) -> str:
    tks = family.split(".")
    if len(tks) <= 1 or len(tks) >= 3:
        return family

    return tks[0]


def family_to_minor(family: str) -> str:
    tks = family.split(".")
    if len(tks) <= 1 or len(tks) >= 3:
        return family

    return tks[1]


def tag_to_patch(tag: str) -> str:
    tks = tag_to_version_str(tag).split(".")
    if len(tks) != 4:
        return "0"

    return tks[3]


def tag_to_improv(tag: str) -> str:
    tks = tag_to_version_str(tag).split(".")
    if len(tks) != 4:
        return "0"

    return tks[2]


def version_to_family(version: str) -> str:
    return tag_to_family(version)


def version_str_to_int(version: str) -> int:
    v = tag_to_version_str(version)

    tks = v.split(".")
    tks_count = len(tks)
    if tks_count > 4 or tks_count <= 2:
        return -1

    version_as_int = 0
    for i in range(tks_count):
        tk = tks[i]
        tk_val = 0
        try:
            tk_val = int(tk)
        except ValueError:
            return -1
        version_as_int += tk_val * (1000 ** (3 - i))

    return version_as_int


def version_int_to_str(version_as_int: int) -> str:

    major_i: int = int(version_as_int/(1000 ** 3)) % 1000
    minor_i: int = int(version_as_int/(1000 ** 2)) % 1000
    improv_i: int = int(version_as_int/1000) % 1000
    patch_i: int = int(version_as_int) % 1000

    if major_i <= 3:
        version_str = "{}.{}.{}.{}".format(major_i, minor_i, improv_i, patch_i)
    else:
        version_str = "{}.{:02}.{:02}.{:02}".format(major_i, minor_i, improv_i, patch_i)
    return version_str


def family_str_to_int(family: str) -> int:
    fam = tag_to_family(family)

    tks = fam.split(".")
    tks_count = len(tks)
    if tks_count > 3:
        return -1

    family_as_int = 0
    for i in range(tks_count):
        tk = tks[i]
        tk_val = 0
        try:
            tk_val = int(tk)
        except ValueError:
            return -1

        family_as_int += tk_val * (1000 ** (3 - i))

    return family_as_int


def tag_to_version(tag: str) -> Version:
    v = Version()
    v.as_str = tag_to_version_str(tag)
    v.family = tag_to_family(tag)
    v.major = family_to_major(v.family)
    v.minor = family_to_minor(v.family)
    v.improv = tag_to_improv(v.as_str)
    v.patch = tag_to_patch(v.as_str)
    v.as_int = version_str_to_int(v.as_str)
    v.as_ags4_str = version_int_to_str(v.as_int)
    v.family_as_int = family_str_to_int(v.family)
    return v


def version_str_to_version(version_str: str) -> Version:
    return tag_to_version(version_str)



from __future__ import annotations  # for python 3.8

import os
from subprocess import Popen

from agstoolbox.core.ags.ags_editor import LocalAgsEditor
from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.utils.file import get_absolute_path
from agstoolbox.core.utils.run import run_exe_params


def ags_editor_start(editor: LocalAgsEditor, block: bool = False) -> int:
    return run_exe_params(editor.path, block)


def ags_editor_load(editor: LocalAgsEditor, project: GameProject, block: bool = False) -> int:
    project_path = get_absolute_path(project.path)
    return run_exe_params(editor.path, block, [project_path], )


def ags_editor_build(editor: LocalAgsEditor, project: GameProject, block: bool = False) -> int:
    project_path = get_absolute_path(project.path)
    return run_exe_params(editor.path, block, ['/compile', project_path])




from codecs import open
from pathlib import Path

from setuptools import setup, find_packages
import re

with open('src/agstoolbox/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()

setup(
    name='agstoolbox',
    version=version,
    description='Utility to help manage Adventure Game Studio Projects and Editors.',
    long_description=long_description,
    url='https://github.com/ericoporto/agstoolbox',
    download_url='https://github.com/ericoporto/agstoolbox/tarball/' + version,
    author='erico',
    author_email='eri0onpm@gmail.com',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only'
    ],
    keywords=[
        'AGS Toolbox', 'Adventure Game Studio', 'development', 'ags', 'Game Development', 'gamedev'
    ],
    python_requires='>=3.8',
    install_requires=[
        'pyqt6',
        'requests',
        'defusedxml',
        'platformdirs',
        'pefile',
        'shtab'
    ],
    packages=[
        'agstoolbox',
        'agstoolbox.core',
        'agstoolbox.core.ags',
        'agstoolbox.core.cmdline',
        'agstoolbox.core.gh',
        'agstoolbox.core.settings',
        'agstoolbox.core.utils',
        'agstoolbox.core.version',
        'agstoolbox.panels',
        'agstoolbox.system',
        'agstoolbox.wdgts',
        'agstoolbox.wdgts_utils'
    ],
    package_dir={"": "src"},
    scripts=["agstoolbox", "atbx"],
    package_data={
        'agstoolbox': ['data/*.png', 'data/fonts/*.ttf']
    },
)



from __future__ import annotations  # for python 3.8
from PyQt6 import QtCore, QtGui, QtWidgets
from PyQt6.QtWidgets import QTreeWidgetItem, QWidget, QLabel, QHBoxLayout, QVBoxLayout, QGridLayout
from PyQt6.QtGui import QAction

from agstoolbox.at_icons import main_icon_as_pixmap
from agstoolbox.core.ags.ags_editor import LocalAgsEditor
from agstoolbox.core.ags.ags_local_run import ags_editor_load
from agstoolbox.wdgts_utils.ags_local_extra import ags_project_folder_in_explorer
from agstoolbox.core.ags.game_project import GameProject
from agstoolbox.core.utils.time import s_ago
from agstoolbox.wdgts_utils.action_utils import DefaultMenuQAction


class ActionEditorPair:
    action: QAction = None
    editor: LocalAgsEditor = None

    def __init__(self, action: QAction = None, editor: LocalAgsEditor = None):
        self.action = action
        self.editor = editor


class ProjectImage(QLabel):
    def __init__(self, img_icon_path: str, parent: QWidget = None):
        QLabel.__init__(self, parent)

        pxmap = main_icon_as_pixmap()
        if img_icon_path:
            pxmap = QtGui.QPixmap(img_icon_path)

        self.setPixmap(pxmap.scaled(
            32, 32,
            QtCore.Qt.AspectRatioMode.KeepAspectRatio,
            QtCore.Qt.TransformationMode.SmoothTransformation))
        self.setMaximumSize(32, 32)
        self.setMinimumSize(32, 32)


class ProjectWidget(QWidget):
    project = None

    def __init__(self, ags_game_project: GameProject, parent: QWidget = None):
        QWidget.__init__(self, parent)
        self.project = ags_game_project

        self.icon_img = ProjectImage(ags_game_project.ico_path)
        self.labelName = QLabel(self.project.name)
        self.labelName.setWordWrap(True)

        smaller_font = QtGui.QFont(
            self.labelName.font().family(),
            self.labelName.font().pointSize() * 0.90,
        )

        smallest_font = QtGui.QFont(
            self.labelName.font().family(),
            self.labelName.font().pointSize() * 0.75,
        )

        self.labelVersion = QLabel(self.project.ags_editor_version.as_str)
        self.labelVersion.setFont(smaller_font)
        self.labelVersion.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self.labelTime = QLabel(s_ago(self.project.last_modified))
        self.labelTime.setFont(smaller_font)
        self.labelTime.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

        self.labelDir = QLabel(self.project.directory)
        self.labelDir.setFont(smallest_font)
        self.labelDir.setDisabled(True)

        top_hbox = QHBoxLayout()
        bottom_hbox = QHBoxLayout()

        right_vbox = QVBoxLayout()

        vbox = QVBoxLayout()

        vbox.addLayout(top_hbox)
        vbox.addLayout(bottom_hbox)
        top_hbox.addWidget(self.labelName)
        right_vbox.addWidget(self.labelTime)
        right_vbox.addWidget(self.labelVersion)
        top_hbox.addLayout(right_vbox)
        bottom_hbox.addWidget(self.labelDir)

        main_qgrid = QGridLayout()
        main_qgrid.addWidget(self.icon_img)
        main_qgrid.addLayout(vbox, 0, 1)

        self.setLayout(main_qgrid)

    def quick_open_project(self):
        self.parent().parent().tools_tree.open_project_tool(self.project)

    def quick_build_project(self):
        self.parent().parent().tools_tree.build_project_tool(self.project)

    def mouseDoubleClickEvent(self, event):
        self.quick_open_project()

    def get_managed_editors(self) -> list[LocalAgsEditor]:
        return self.parent().parent().tools_tree.managed_editors_list

    def get_unmanaged_editors(self) -> list[LocalAgsEditor]:
        return self.parent().parent().tools_tree.unmanaged_editors_list

    def set_managed_editors_menu(self,
                                 parent_menu: QtWidgets.QMenu = None) -> list[ActionEditorPair]:
        submenu = QtWidgets.QMenu("Open in Managed Editor", parent_menu)
        parent_menu.addMenu(submenu)
        editors = self.get_managed_editors()
        actions: list[ActionEditorPair] = list()
        for editor in editors:
            action = submenu.addAction(editor.name)
            actions.append(ActionEditorPair(action=action, editor=editor))

        return actions

    def set_unmanaged_editors_menu(self,
                                   parent_menu: QtWidgets.QMenu = None) -> list[ActionEditorPair]:
        submenu = QtWidgets.QMenu("Open in External Editor", parent_menu)
        parent_menu.addMenu(submenu)
        editors = self.get_unmanaged_editors()
        actions: list[ActionEditorPair] = list()
        for editor in editors:
            action = submenu.addAction(editor.name)
            actions.append(ActionEditorPair(action=action, editor=editor))

        return actions

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu(self)
        quick_open_action = DefaultMenuQAction(menu, "Quick Open Project")
        quick_build_action = menu.addAction("Quick Build Project")
        open_folder_action = menu.addAction("Open Folder in File Explorer")
        menu.addSeparator()
        managed_actions = self.set_managed_editors_menu(menu)
        unmanaged_actions = self.set_unmanaged_editors_menu(menu)

        action = menu.exec(self.mapToGlobal(event.pos()))
        if action == open_folder_action:
            ags_project_folder_in_explorer(self.project)
            return
        elif action == quick_open_action:
            self.quick_open_project()
            return
        elif action == quick_build_action:
            self.quick_build_project()
            return
        else:
            for a_pair in managed_actions:
                if a_pair.action == action:
                    ags_editor_load(a_pair.editor, self.project)
                    return

            for a_pair in unmanaged_actions:
                if a_pair.action == action:
                    ags_editor_load(a_pair.editor, self.project)
                    return


class TreeItemProject(QTreeWidgetItem):
    itm_wdgt = None

    def __init__(self, ags_game_project: GameProject):
        QTreeWidgetItem.__init__(self)
        self.itm_wdgt = ProjectWidget(ags_game_project)

    def updateInTree(self):
        self.treeWidget().setItemWidget(self, 0, self.itm_wdgt)
